/* global log, oscObjects */

class Config {
  constructor(
    serverCredentials,
    serverUrl,
    agentLanguage,
    languageFieldColumnName,
    languageFieldDisplayName,
    qualityRatingType,

    wordCountThreshold = 0,
    wordCountThresholdMessage = "You've requested a translation of <wordCount> words.\n\nDo you wish to continue or perform partial translation instead?\n\nPress OK to process translation or Cancel to proceed with partial translation.",

    wordCountHardLimit = 0,
    wordCountHardLimitMessage = "You've requested a translation of <wordCount> words, which is above the limit set by your administrator.\n\nPlease perform a partial translation instead."
  ) {
    this.serverCredentials = serverCredentials;
    this.serverUrl = serverUrl;

    this.agentLanguage = parseInt(agentLanguage, 10);
    this.languageFieldColumnName = languageFieldColumnName;
    this.languageFieldDisplayName = languageFieldDisplayName;

    this.translationType = "machine";
    this.retranslationType = "machine2";

    this.qualityRatingInstructions = "";
    this.qualityRatingType = qualityRatingType;
    this._qualityRatingConfigsLoaded = false;

    this.wordCountThreshold = wordCountThreshold;
    this.wordCountThresholdMessage = wordCountThresholdMessage;
    this.wordCountHardLimit = wordCountHardLimit;
    this.wordCountHardLimitMessage = wordCountHardLimitMessage;
  }

  /** @returns {Promise<int>} the id of the current customer language */
  async getCustomerLanguage() {
    return new Promise((resolve) => {
      oscObjects.extensionProvider.registerWorkspaceExtension((workspaceRecord) => {
        const name = "Incident." + this.languageFieldColumnName;
        workspaceRecord.getFieldValues([name]).then((fieldDetails) => {
          const languageId = parseInt(fieldDetails.getField(name).getValue(), 10);
          return resolve(languageId);
        });
      });
    });
  }

  async loadAgentQualityRatingConfig(lioRestManager) {
    if (this._qualityRatingConfigsLoaded) {
      return Promise.resolve();
    }
    return new Promise((resolve, reject) => {
      // TODO: when the server has more rating instructions, request the one pertinent to the configuring rating type
      // ex: `QUALITY_RATING_INSTRUCTIONS_TYPE_${this.qualityRatingType}`,
      lioRestManager
        .getInstruction("TRANSLATION_RATING_LABEL", this.agentLanguage, false, "Flag to report poor translation")
        .then((instructions) => {
          this.qualityRatingInstructions = instructions;
          this._qualityRatingConfigsLoaded = true;
          return resolve();
        })
        .catch((error) => reject(error.desc));
    });
  }
}

/** @returns dict containing the SCPs for this agent */
async function _getServerConfigurationProperties() {
  const configNames = [
    "languageioLogin",
    "languageioUrl",
    "languageField",
    "agentLanguage",
    "qualityRatingType",
    "wordcountThreshold",
    "wordcountAlertMessage",
    "wordcountHardLimit",
    "wordcountHardLimitAlertMessage",
  ];
  return new Promise((resolve, reject) => {
    oscObjects.globalContext
      .getExtensionContext("LanguageIO_BUI_Incident")
      .then((extensionContext) => {
        extensionContext
          .getProperties(configNames)
          .then((collection) => {
            const configs = [];
            configNames.forEach((name) => {
              const config = collection.get(name);
              if (config) {
                configs[name] = config.getValue();
              }
            });
            return resolve(configs);
          })
          .catch((error) => reject(error.desc));
      })
      .catch((error) => reject(error.desc));
  });
}

// pulled out into separate function because we need to do async work to determine all the values
/* eslint-disable-next-line no-unused-vars */ // used in main to create config object
async function createConfig(oscRestManager) {
  let configs;
  try {
    configs = await _getServerConfigurationProperties();
  } catch (error) {
    log.error(`Could not retrieve server configuration properties: ${error}`);
    return Promise.reject(
      new Error(
        "Failed to retrieve Server Configuration Properties. Please view the browser's developer tools console or contact your Language I/O representative for assistance."
      )
    );
  }

  if (configs.languageioLogin === null || configs.languageioLogin.length < 1) {
    return Promise.reject(
      new Error(
        "The `languageioLogin` Server Configuration Property is not set up for your profile. Please contact your Language I/O representative for assistance."
      )
    );
  }

  let languageFieldColumnName = "lang_id";
  let languageFieldDisplayName = "Language";
  if (configs.languageField !== "default") {
    try {
      languageFieldColumnName = "C$" + (await oscRestManager.getCustomLanguageLookupName(configs.languageField));
      languageFieldDisplayName = configs.languageField;
    } catch (error) {
      return Promise.reject(error);
    }
  }

  return Promise.resolve(
    new Config(
      configs.languageioLogin,
      configs.languageioUrl,
      configs.agentLanguage,
      languageFieldColumnName,
      languageFieldDisplayName,
      configs.qualityRatingType,
      // force `undefined` instead of null or empty strings so the default parameter values are used
      Math.abs(configs.wordcountThreshold) || undefined,
      configs.wordcountAlertMessage || undefined,
      Math.abs(configs.wordcountHardLimit) || undefined,
      configs.wordcountHardLimitAlertMessage || undefined
    )
  );
}
