/* global log, tagVariables */

/* eslint-disable-next-line no-unused-vars */ // used in other files
class LioRestManager {
  constructor(config, oscRestManager) {
    this._incidentId = oscRestManager.incidentId;
    this._config = config;
    this._headers = {
      "Content-Type": "application/json",
      Authorization: "Basic " + this._config.serverCredentials,
    };
    this._oscRestManager = oscRestManager;
  }

  /**
   * Gets custom instruction text for the given instruction type and locale
   * @param {string} instructionType the name of the instruction
   * @param {string} consumerLocaleId the identifier of the consumer locale
   * @param {boolean} [disableDefault] (optional) sets the `disableDefault` GET param
   * @param {string} [defaultInstructionContent] (optional) default instruction text if server does not return content
   * @returns {Promise<string>} custom instruction text, if it exists
   */
  async getInstruction(instructionType, consumerLocaleId, disableDefault = false, defaultInstructionContent = "") {
    const fullUrl = `${
      this._config.serverUrl
    }/linguistnow/resources/com.linguistnow.entities.consumerhasinstruction/getConsumerHasInstruction/?consumerLocaleId=${consumerLocaleId}&disableDefault=${
      disableDefault ? "true" : "false"
    }&instructionType=${instructionType}`;
    const parameters = {
      method: "GET",
      headers: this._headers,
    };
    return new Promise((resolve, reject) => {
      fetch(fullUrl, parameters)
        .then((response) => {
          if (response.ok) {
            response
              .json()
              .then((instruction) => {
                if (instruction.instructionContent) {
                  resolve(instruction.instructionContent);
                } else if (defaultInstructionContent !== "") {
                  resolve(defaultInstructionContent);
                } else {
                  reject(new Error("No localized value for the instruction was returned."));
                }
              })
              .catch((error) => reject(error));
          } else {
            reject(response);
          }
        })
        .catch((error) => reject(error));
    });
  }
  async autoDetectLocale(payload){
    const fullUrl = `${this._config.serverUrl}/linguistnow/resources/locale/detectLocale`;
    const method = "POST";
    const parameters = {
      method,
      headers: this._headers,
      body: JSON.stringify(payload),
    };
    return new Promise((resolve, reject) => {
      fetch(fullUrl, parameters)
        .then((response) => {
          if (response.ok) {
            response
              .json()
              .then((detectedLanguage) => {
                if (detectedLanguage.error) {
                  reject(new Error(detectedLanguage.error.userMessage));
                } else {
                  resolve(Number(detectedLanguage.detectedLanguage));
                }
              })
              .catch((error) => reject(error));
          } else {
            reject(response);
          }
        })
        .catch((error) => reject(error));
    });
  }

  async getTranslation(fromAgent, retranslation, partial, text, sourceThreadId = 0, privateNote = false) {
    if (text.length < 1) {
      throw new Error("Translation cancelled: There is no content to translate.");
    }

    /* agent editor content has HTML formatting, and OSC standard content has already had its
    formatting fixed, so at this point we just have to make sure any variables are marked
    no-translate */
    const fixedText = tagVariables(text);
    const translationEligibilityTypeIds = [];
    var customerLanguage = await this._config.getCustomerLanguage();
    const needsDetected = isNaN(customerLanguage)

    if (needsDetected){
      const autoDetectRequest = {
        text: fixedText,
        translationType: this._config.translationType
      }
      if(!fromAgent){
        customerLanguage = await this.autoDetectLocale(autoDetectRequest);
        await this._oscRestManager.updateLanguageField(this._config.languageFieldColumnName, customerLanguage);
      }
      else{
        throw new Error(
          "Customer language not yet detected.\nTranslate a customer message or manually set the language field" 
        );
      }
    }

    if (fromAgent) {
      // for agent messages, we create an LN Thread with the original content before translating
      return this._oscRestManager
        .createLNThread(fixedText, 1, sourceThreadId, 0, " ", 0, 0, 0, 0, 0, 0, `${this._config.agentLanguage}`)
        .catch(() => {
          throw new Error(
            "Translation failed: Couldn't create LNThread to store the original agent message.\nVerify that your profile has LNThread permissions."
          );
        })
        .then((untranslatedThread) => this._translate(
            fixedText,
            this._config.agentLanguage,
            customerLanguage,
            this._config.translationType,
            0,
            untranslatedThread.id,
            2,
            translationEligibilityTypeIds
          )
        );
    } else {
      let translationType = this._config.translationType;
      const retranslationType = this._config.retranslationType;
      let completedThreadType = 4;
      if (privateNote) {
        if (partial) {
          completedThreadType = 8;
        }
        else {
          completedThreadType = 7
        }
      } else if (partial) {
        completedThreadType = 6;
      } else if (retranslation){
        translationType = retranslationType;
        completedThreadType = 5;
      }
      if (!retranslation) {
        switch (retranslationType) {
          case "human":
            translationEligibilityTypeIds.push(3);
            break;
          case "machine2":
            translationEligibilityTypeIds.push(2);
            break;
          default:
            log.warn('Invalid value "' + translationType + '" for config "retranslationType"');
            break;
        }
      }
      if (this._config.qualityRatingType > 0) {
        translationEligibilityTypeIds.push(1);
      }
      return this._translate(
        fixedText,
        customerLanguage,
        this._config.agentLanguage,
        translationType,
        sourceThreadId,
        0,
        completedThreadType,
        translationEligibilityTypeIds
      );
    }
  }

  /**
   * creates a rating for a translation (**by the translation's ID**)
   * @see {@link https://languageio.atlassian.net/wiki/spaces/DEV/pages/2024112146/LIO+Rest+API+for+Rating}
   * @param {number} payload.translationId the ID of the LIO translation
   * @param {number} payload.ratingTypeId the type of the rating
   * @param {number} payload.ratingPoints the points/value of the rating
   * @returns
   */
  async createTranslationRating(payload) {
    return this._rateTranslation("POST", payload);
  }

  /**
   * updates the rating for a translation (**by the existing/previous rating's ID**)
   * @see {@link https://languageio.atlassian.net/wiki/spaces/DEV/pages/2024112146/LIO+Rest+API+for+Rating}
   * @param {number} payload.ratingId the ID of the LIO rating
   * @param {number} payload.ratingTypeId the type of the rating
   * @param {number} payload.ratingPoints the points/value of the rating
   * @returns
   */
  async updateTranslationRating(payload) {
    return this._rateTranslation("PUT", payload);
  }

  /**
   * wrapper for API requests to create or update a translation's rating
   * @see {@link https://languageio.atlassian.net/wiki/spaces/DEV/pages/2024112146/LIO+Rest+API+for+Rating}
   * @param {string} method the HTTP method for the fetch request
   * @param {number} [payload.translationId] the ID of the LIO translation - required for new ratings only
   * @param {number} [payload.ratingId] the ID of the LIO rating - required for existing ratings only
   * @param {number} payload.ratingTypeId the type of the rating
   * @param {number} payload.ratingPoints the points/value of the rating
   * @returns {object} the rating data (if successful)
   */
  async _rateTranslation(method, payload) {
    const fullUrl = `${this._config.serverUrl}/linguistnow/resources/v1/rating`;
    const parameters = {
      method,
      headers: this._headers,
      body: JSON.stringify(payload),
    };
    return new Promise((resolve, reject) => {
      fetch(fullUrl, parameters)
        .then((response) => {
          if (response.ok) {
            response
              .json()
              .then((rating) => {
                if (rating.error) {
                  reject(new Error(rating.error.userMessage));
                } else {
                  resolve(rating);
                }
              })
              .catch((error) => reject(error));
          } else {
            reject(response);
          }
        })
        .catch((error) => reject(error));
    });
  }

  _translate(
    text,
    sourceLocale,
    targetLocale,
    translationType,
    sourceThreadId,
    sourceLNThreadId,
    completedThreadType,
    translationEligibilityTypeIds = []
  ) {
    const fullUrl = this._config.serverUrl + "/linguistnow/resources/v1/translate";
    const requestBody = {
      sourceContent: text,
      sourceLocale,
      targetLocale,
      contentTypeName: "ticket",
      translationType,
      incidentId: this._incidentId,
      consumerSourceId: sourceThreadId,
      textType: "html",
      crmThreadType: completedThreadType,
      translationEligibilityTypeIds,
    };
    const parameters = {
      method: "POST",
      headers: this._headers,
      body: JSON.stringify(requestBody),
    };

    return (
      fetch(fullUrl, parameters)
        .then((response) => {
          if (response.ok) {
            return response.json().then((translation) => ({ translation }));
          } else {
            // if !response.ok, we authenticated THEN recieved an error
            return response.json().then((error) => ({ error }));
          }
        })
        /* either didn't authenticate successfully, or there was a problem with the response body */
        .catch((error) => ({ error }))
        .then(({ translation, error }) => {
          if (translation) {
            this._rnResponseProjectId = translation.rnResponseProjectId;
            const successMessage = "SUCCESS: " + translation.statusMessage;
            this._oscRestManager.updateStatuses(successMessage, sourceLNThreadId, successMessage);

            // for human translation, the server makes the LNThread later
            if (translationType === "human") {
              return Promise.resolve();
            }

            // for machine translation, the extension makes the LNThread now
            return this._oscRestManager
              .createLNThread(
                translation.translatedText,
                completedThreadType,
                sourceThreadId,
                sourceLNThreadId,
                successMessage,
                this._rnResponseProjectId,
                translation.wordCount,
                translation.translationId,
                0,0,0,  //Default values for ratingId, ratingTypeId, and ratingPoints
                `${sourceLocale}`,
                `${targetLocale}`
              )
              .catch(() => {
                throw new Error(
                  "The translation succeeded, but we couldn't create an LNThread to store the translation.\nVerify that your profile has LNThread permissions."
                );
              });
          } else if (error && error.error) {
            /* error.error is a server error, which will always have
          an id, userMessage, and message */
            return this._oscRestManager
              .updateStatuses("Translation failed.", sourceLNThreadId, "ERROR: " + error.error.userMessage)
              .then(() => {
                throw new Error("Translation failed: " + error.error.message);
              });
          } else {
            const shortError = "ERROR: Translation failed.";
            const longError =
              "Translation failed: Could not access Language I/O server. Verify your server URL and credentials and try again.";
            return this._oscRestManager.updateStatuses(shortError, sourceLNThreadId, "ERROR: " + longError).then(() => {
              throw new Error(longError);
            });
          }
        })
    );
  }
}
