/* global ORACLE_SERVICE_CLOUD, createConfig, LioRestManager, OscRestManager, Jodit, Log,
ThreadGroup, stripHtmlTags, getWordCount */

/* GLOBAL VARIABLES */
let log;
let allThreads = [];
let untranslatedThreads = [];
let joditEditor;

/* asyncs for handling program-wide tasks like REST calls, loaded in runExtension */
let oscRestManager;
let lioRestManager;
let profileConfig;

/** asynchronously-retrieved objects from the OSC BUI API, loaded in runExtension.
 *
 * Oracle uses their own version of Promises that behave slightly different than standard
 * JS Promises, and their async functions may need to be wrapped in standard Promises to avoid
 * unexpected execution order.
 *
 * Oracle's promises also do not support .finally()
 */
const oscObjects = {
  extensionProvider: null,
  globalContext: null,
};

/* INITIALIZE THE EXTENSION */

/* eslint-disable-next-line no-unused-vars */ // used by init.html to initialize extension
async function runExtension() {
  displayLoadingText();

  // docs aren't clear about how the OSC functions might fail, wrapping it all in try as a backup
  try {
    // enter the main bui framework function where we can get the site & incident info we need
    /* had to nest with multiple catches instead of flat chaining because the OSC calls weren't
    behaving the same way as standard promises */
    oscObjects.extensionProvider = await ORACLE_SERVICE_CLOUD.extension_loader.load("Language I/O Incident");
    oscObjects.globalContext = await oscObjects.extensionProvider.getGlobalContext();

    log = new Log(oscObjects.extensionProvider.getLogger());
    oscObjects.extensionProvider.registerWorkspaceExtension((workspaceRecord) => {
      const incidentId = workspaceRecord.getWorkspaceRecordId();
      /* we may be trying to auto-load an unsaved incident. we can't make the necessary rest
      calls until the incident has been saved, so if it hasn't, stop and show a load button. */
      if (incidentId < 1) {
        log.warn(new Error("Can't load Language I/O. Save the incident and try again."));
        displayPreloadedView();
      } else {
        oscObjects.globalContext.getSessionToken().then((sessionToken) => {
          // store all of the info we need from the BUI framework to make our REST calls
          oscRestManager = new OscRestManager(sessionToken, incidentId);
          createConfig(oscRestManager).then((config) => {
            profileConfig = config;
            lioRestManager = new LioRestManager(profileConfig, oscRestManager);
            profileConfig.loadAgentQualityRatingConfig(lioRestManager).then(() => {
              loadDisplay();
            });
          });
        });
      }
    });
  } catch (error) {
    displayPreloadedView(error);
  }
}

/* LOADING AND REFRESHING DISPLAY */

function showLoadingStyle() {
  document.body.classList.add("loading");
}

function hideLoadingStyle() {
  document.body.classList.remove("loading");
}

function displayPreloadedView(error) {
  document.getElementById("agent_input").innerHTML = "";
  document.getElementById("thread_list").innerHTML = "";
  document.getElementById("load_button").style.display = "flex";
  if (error) {
    log.error(error);
    Log.alert("Could not load Language I/O: \n" + error);
  }
  hideLoadingStyle();
}

function displayLoadingText() {
  showLoadingStyle();
  document.getElementById("load_button").style.display = "none";
  document.getElementById("agent_input").innerHTML = '<div class="load_text">Loading text editor...</div>';
  document.getElementById("thread_list").innerHTML = '<div class="load_text">Loading thread list...</div>';
}

function loadDisplay() {
  document.getElementById("load_button").style.display = "none";
  displayTopBarButtons();
  displayAgentTextEntry();
  refreshThreadDisplay();
  hideLoadingStyle();
}

function displayTopBarButtons() {
  const topButtons = document.getElementById("extension_buttons");
  topButtons.innerHTML =
    '<button id="agent_input_button" onclick="translateContent(true, false, escape(joditEditor.getEditorValue()), 0);">Translate Input</button>' +
    '<button id="refresh_button" onclick="refreshExtension();">Refresh Threads</button>' +
    `<button id="translate_all_threads_button" onclick="translateAllThreads()" title="Translate all threads" class="hidden"></button>` +
    "<hr />";
}

function displayAgentTextEntry() {
  const agentInput = document.getElementById("agent_input");
  agentInput.innerHTML = '<input id="jodit_input" type="text" value=""/>';
  joditEditor = new Jodit("#jodit_input", {
    askBeforePasteHTML: false,
    askBeforePasteFromWord: false,
    defaultActionOnPaste: "insert_clear_html",
    allowResizeY: true,
    showCharsCounter: false,
    showWordsCounter: false,
    showXPathInStatusbar: false,
    link: {
      followOnDblClick: false,
      processVideoLink: false,
      noFollowCheckbox: false,
      openInNewTabCheckbox: false,
    },
    popup: {
      // reference: https://github.com/xdan/jodit/blob/3.3.8/src/plugins/inline-popup.ts#L37-L59
      // intent is to remove the Brush and File buttons from the inline popup shown on anchor tags
      a: Jodit.Array([
        {
          name: "eye",
          tooltip: "Open link",
          exec: (editor, current) => {
            if (current) {
              const href = current.getAttribute("href");
              if (href) {
                editor.ownerWindow.open(href);
              }
            }
          },
        },
        {
          name: "link",
          tooltip: "Edit link",
          icon: "pencil",
        },
        "unlink",
        // "brush",
        // "file"
      ]),
    },
    uploader: {
      insertImageAsBase64URI: false,
    },
    enter: "div",
    enterBlock: "div",
    toolbarAdaptive: false,
    buttons: [
      "undo",
      "redo",
      "|",
      "font",
      "|",
      "fontsize",
      "|",
      "bold",
      "italic",
      "underline",
      "strikethrough",
      "|",
      "brush",
      "|",
      "align",
      "|",
      "ul",
      "ol",
      "|",
      "link",
      "|",
      "subscript",
      "superscript",
      "|",
      "image",
    ],
  });
  joditEditor.value = "";
  agentInput.addEventListener("keyup", displayStandardTextPopup);
}

function refreshThreadDisplay() {
  return oscRestManager
    .loadAllThreads()
    .then((freshThreads) => {
      allThreads = freshThreads;
      updateTranslateAllThreadsButton()
      displayLoadedThreads();
    })
    .catch((error) => {
      const modError = error;
      modError.message = "Couldn't refresh threads: " + modError.message;
      log.warn(modError);
    });
}

function updateTranslateAllThreadsButton(){
  untranslatedThreads = getAllUntranslatedThreads(`${profileConfig.agentLanguage}`);
  const translateAllThreadsButton = document.getElementById("translate_all_threads_button");
  if (untranslatedThreads.length > 0) {
    translateAllThreadsButton.innerText = `Translate ${untranslatedThreads.length} Threads`;
    translateAllThreadsButton.classList.remove("hidden");
  } else {
    translateAllThreadsButton.classList.add("hidden");
  }
}

function displayLoadedThreads() {
  // this whole function is gross, but it works - will get refactored when switching to react

  // for now, hardcoded to the BUI default of most recent first (except orphans)
  allThreads.sort((a, b) => Date.parse(b.createdTime) - Date.parse(a.createdTime));

  const threadGroups = [];
  const uncategorizedThreads = [];
  Object.keys(allThreads).forEach((key) => {
    uncategorizedThreads.push(allThreads[key]);
  });

  // get all threads of parent-only types, or with no parent, and create a ThreadGroup for each
  Object.keys(allThreads).forEach((key) => {
    const currentThread = allThreads[key];
    if (
      currentThread.hierarchy === "parent" ||
      (currentThread.sourceSystemThread === 0 && currentThread.sourceLNThread === 0)
    ) {
      const currentGroup = new ThreadGroup(currentThread);

      threadGroups.push(currentGroup);
      uncategorizedThreads.splice(uncategorizedThreads.indexOf(currentThread), 1);
    }
  });

  // treat all children as orphans until we find their parents
  const orphanThreads = [];
  Object.keys(uncategorizedThreads).forEach((key) => {
    orphanThreads.push(uncategorizedThreads[key]);
  });

  // if the child has a parent, add them to that parents' group, and remove from orphan list
  Object.keys(uncategorizedThreads).forEach((key) => {
    const currentThread = uncategorizedThreads[key];
    let parentThreadId = 0;

    /* for agent translation nesting we care about the untranslated agent message - an LNThread
    for other child thread nesting we care about the untranslated customer message - a sys thread */
    if (currentThread.threadType === 2 && currentThread.sourceLNThread !== 0) {
      parentThreadId = currentThread.sourceLNThread;
    } else if (currentThread.sourceSystemThread !== 0) {
      parentThreadId = currentThread.sourceSystemThread;
    }

    Object.keys(threadGroups).forEach((parentKey) => {
      const currentGroup = threadGroups[parentKey];
      if (parseInt(currentGroup.parent.id, 10) === parseInt(parentThreadId, 10)) {
        currentGroup.children.push(currentThread);
        orphanThreads.splice(orphanThreads.indexOf(currentThread), 1);
      }
    });
  });

  // get everyone's html, letting any orphans bubble to the top for now
  // (pending server code update, human translations of agent messages are always orphans)
  let formattedThreads = "";
  Object.keys(orphanThreads).forEach((key) => {
    formattedThreads += "<p>" + orphanThreads[key].getHtml() + "</p>";
  });
  Object.keys(threadGroups).forEach((key) => {
    formattedThreads += threadGroups[key].getGroupHtml();
  });
  const threadsDiv = document.getElementById("thread_list");
  threadsDiv.innerHTML = "<hr />" + formattedThreads + "<hr />";

  // set all links in threads to open in a new tab to avoid redirecting away from the extension
  const links = threadsDiv.getElementsByTagName("a");
  Object.keys(links).forEach((id) => {
    links[id].target = "_blank";
  });

  Object.keys(threadGroups).forEach((key) => {
    threadGroups[key].makeCollapsable();
  });
}

/* BUTTON EVENTS */

/* eslint-disable-next-line no-unused-vars */ // used by the refresh extension button
function refreshExtension() {
  showLoadingStyle();
  const refreshButton = document.getElementById("refresh_button");
  refreshButton.disabled = true;
  refreshButton.innerHTML = "Refreshing Threads...";

  refreshThreadDisplay()
    .catch((error) => {
      const modError = error;
      modError.message = "Could not refresh Language I/O: \n" + modError.message;
      log.error(modError);
      Log.alert(modError);
    })
    .finally(() => {
      refreshButton.disabled = false;
      refreshButton.innerHTML = "Refresh Threads";
      hideLoadingStyle();
    });
}

/* eslint-disable-next-line no-unused-vars */ // used by status buttons
function toggleStatus(statusId) {
  const fullStatus = document.getElementById(statusId);
  if (fullStatus.classList.contains("collapsed")) {
    fullStatus.classList.remove("collapsed");
  } else {
    fullStatus.classList.add("collapsed");
  }
}

/* eslint-disable-next-line no-unused-vars */ // used in thread (re)translate buttons
function translateContent(
  fromAgent,
  retranslation,
  text,
  sourceSystemThread = 0,
  sourceButtonId = "",
  privateNote = false
) {
  let partial = false;
  let textToTranslate = unescape(text);
  if (!retranslation) {
    // no support for partial retranslations
    try {
      // check if any content is highlighted, and if so do partial translation on just that.
      const selection = window.getSelection();
      if (selection !== null && selection.type === "Range" && selection.rangeCount > 0) {
        const wrapper = document.createElement("div");
        for (let i = 0; i < selection.rangeCount; i += 1) {
          wrapper.appendChild(selection.getRangeAt(i).cloneContents());
        }
        if (wrapper.textContent.length && stripHtmlTags(textToTranslate).includes(wrapper.textContent)) {
          textToTranslate = wrapper.innerHTML;
          partial = true;
        }
      }
    } catch (e) {
      // assume nothing is selected, and move forward with full translation
    }
  }

  if (!retranslation && !partial && !fromAgent) {
    if (profileConfig.wordCountThreshold > 0 || profileConfig.wordCountHardLimit > 0) {
      const estimatedWordCount = getWordCount(stripHtmlTags(textToTranslate));
      if (estimatedWordCount > profileConfig.wordCountHardLimit && profileConfig.wordCountHardLimit > 0) {
        // estimated word count is above the "hard" limit; alert the agent and do NOT let them proceed with translation at all
        // eslint-disable-next-line no-alert
        alert(profileConfig.wordCountHardLimitMessage.replace(/<wordCount>/gi, estimatedWordCount));
        return;
      } else if (estimatedWordCount > profileConfig.wordCountThreshold && profileConfig.wordCountThreshold > 0) {
        // estimated word count is above the "soft" threshold; prompt the agent if they want to proceed with translation
        // eslint-disable-next-line no-alert, no-restricted-globals
        if (!confirm(profileConfig.wordCountThresholdMessage.replace(/<wordCount>/gi, estimatedWordCount))) {
          // agent clicked "cancel", indicating they do NOT want to proceed with translation
          return;
        }
      }
    }
  }

  let translateButton = document.getElementById("agent_input_button");
  let loadingText = "Translating Input...";
  let finalText = "Translate Input";

  if (!fromAgent) {
    translateButton = document.getElementById(sourceButtonId);
    if (retranslation) {
      loadingText = "Retranslating...";
      finalText = "Retranslate";
    } else {
      loadingText = "Translating...";
      finalText = "Translate";
    }
  }
  showLoadingStyle();
  translateButton.disabled = true;
  translateButton.innerHTML = loadingText;

  lioRestManager
    .getTranslation(fromAgent, retranslation, partial, textToTranslate, sourceSystemThread, privateNote)
    .then(() => {
      if (fromAgent) {
        joditEditor.value = "";
      }
    })
    .catch((error) => {
      log.error(error);
      Log.alert(error);
    })
    .finally(() => {
      refreshThreadDisplay().then(() => {
        translateButton.disabled = false;
        translateButton.innerHTML = finalText;
        hideLoadingStyle();
      });
    });
}

/* eslint-disable-next-line no-unused-vars */ // used by translate transcript button
async function translateAllThreads() {
  const button = document.getElementById("translate_all_threads_button");
  button.setAttribute("disabled", "disabled");
  button.classList.add("processing");
  button.innerText = "Translating threads...";

  try {
    const threadsForTranslation = Array.from(untranslatedThreads);
    // attempts to translate each thread and creates an array of successful translation results
    const translationResponses = await Promise.allSettled(

      threadsForTranslation.map((thread) => {
        let privateNote = false;
        if (
          thread instanceof TranslatedNoteThread ||
          thread instanceof NoteThread ||
          thread instanceof SentAgentResponseThread
        ) {
          privateNote = true;
        }
        return lioRestManager.getTranslation(false, false, false, thread.text, thread.id, privateNote);
      })
    ).then((results) => results.map((result) => (result.status === "fulfilled" ? result.value : undefined)).filter(Boolean)
    );

    const failedTranslationsCount = threadsForTranslation.length - translationResponses.length;
    if (failedTranslationsCount > 0) {
      throw new Error(`Failed to translate ${failedTranslationsCount} of ${threadsForTranslation.length} message(s).`);
    }
    untranslatedThreads = [];

  } catch (error) {
    log.error("Failed to translate all threads. " + error);
    Log.alert("Failed to translate all threads.\n\n" + error);
  }

  button.classList.remove("processing");
  button.removeAttribute("disabled");
  refreshThreadDisplay();
}

function getAllUntranslatedThreads(targetLocale) {
  const threadsForTranslation = [];
  // Look through all of the threads
  allThreads.forEach((thread) => {
    // Check if this thread is already in one of our arrays, if it is just break out of loop to the next iteration
    if (threadsForTranslation.includes(thread)) {
      return;
    }
    if (thread instanceof SystemThread) {
      if (thread.threadType === 4 || thread.threadType === 3 || thread.threadType === 2 || thread.threadType === 1) {
        const translatedThread = allThreads.find((t) => t.sourceSystemThread === parseInt(thread.id, 10));
        if (translatedThread === undefined || translatedThread.targetLocale !== targetLocale) {
          threadsForTranslation.push(thread);
        }
      } else {
        // We don't know the language of SentAgentResponse, but can assume it is same as Language field due to lack of autodetect
        threadsForTranslation.push(thread);
      }
    } else if (
      thread instanceof TranslatedAgentThread ||
      thread instanceof UntranslatedAgentThread ||
      thread instanceof UnexpectedLNThread
    ) {
      // We can't relate SentAgentResponse to the Translated/UntranslatedAgentThread(s) so we just assume that they sent them
    } else if (thread.targetLocale !== targetLocale) {
      threadsForTranslation.push(thread);
    }
  });
  return threadsForTranslation;
}

/**
 * toggle the translation's quality rating
 * @param {number} messageId message.id of the message whose translation quality is being rated
 * @param {number} translationId LIO ID of the translation being rated
 * @param {number} [ratingId] optional existing LIO rating ID (for updating/removing a rating)
 * @param {number} [currentRatingPoints] optional current rating points/value
 */
/* eslint-disable-next-line no-unused-vars */ // used by rating flag button
function toggleTranslationFlagRating(threadId, translationId, ratingId, currentRatingPoints) {
  const thread = allThreads.find((t) => t.id === threadId);
  if (!thread) {
    return;
  }

  let ratingFunction = "";
  const ratingRequest = {
    ratingTypeId: 1,
  };

  // eslint-disable-next-line no-param-reassign
  ratingId = parseInt(ratingId, 10);

  if (ratingId > 0) {
    ratingFunction = "updateTranslationRating";
    ratingRequest.ratingId = ratingId;
    ratingRequest.ratingPoints = currentRatingPoints === 1 ? 0 : 1;
  } else {
    ratingFunction = "createTranslationRating";
    ratingRequest.translationId = translationId;
    ratingRequest.ratingPoints = 1;
  }

  thread.setRatingInfo(0, ratingRequest.ratingPoints);
  lioRestManager[ratingFunction](ratingRequest)
    .then((ratingResponse) => {
      thread.setRatingInfo(ratingResponse.ratingId, ratingResponse.ratingPoints);
      oscRestManager._updateLNThreadRating(threadId, ratingResponse);
    })
    .catch(() => {
      thread.setRatingInfo(0, currentRatingPoints);
      oscRestManager._updateLNThreadRating(threadId, { ratingPoints: currentRatingPoints });
    });
}

/* eslint-disable-next-line no-unused-vars */ // used by the agent input button
function sendAgentInput(text) {
  // add the content to an open or new response thread in the BUI message tab
  oscObjects.extensionProvider.registerWorkspaceExtension((incidentWorkspaceRecord) => {
    incidentWorkspaceRecord.getCurrentEditedThread("Response", true).then((threadEntry) => {
      threadEntry.setContentType("html");
      threadEntry.setContent(unescape(text));
      incidentWorkspaceRecord.triggerNamedEvent("switch_to_messages_tab");
    });
  });
}

/* STANDARD TEXT EVENTS */

function displayStandardTextPopup(keyup) {
  if (keyup.key === "F8") {
    let popup = "";
    if (document.getElementById("standard_text_popup")) {
      // if we've already made a standard text popup, we just have to show it
      document.getElementById("standard_text_popup").hidden = false;
      const hotkeyEntry = document.getElementById("standard_text_popup_hotkey_entry");
      hotkeyEntry.value = "";
      hotkeyEntry.hidden = false;
      hotkeyEntry.focus();
    } else {
      // if this is the first time, create it
      popup = document.createElement("div");
      popup.hidden = true;
      document.body.appendChild(popup);
      popup.id = "standard_text_popup";
      // create divs for everything in the popup and hide what we don't need yet
      popup.innerHTML =
        '<div id="standard_text_popup_title_bar">' +
        '<span id="standard_text_popup_title">Standard Text</span>' +
        '<button id="standard_text_popup_close_button" onclick="closeStandardTextPopup()" alt="Close">✖</button>' +
        "</div>" +
        '<input id="standard_text_popup_hotkey_entry" placeholder="Enter Hotkey"></input>' +
        '<div id="standard_text_popup_failure"> Standard Text Item Not Found' +
        '<button id="standard_text_popup_ok_button" onclick="closeStandardTextPopup()"> OK </button>' +
        "</div>";
      const failureText = document.getElementById("standard_text_popup_failure");
      failureText.hidden = true;
      const hotkeyEntry = document.getElementById("standard_text_popup_hotkey_entry");
      // submit request on when agent hits enter
      hotkeyEntry.addEventListener("keyup", (innerKeyup) => {
        if (innerKeyup.key === "Enter") {
          oscRestManager
            .getStandardText(hotkeyEntry.value)
            .then((returnedText) => {
              hotkeyEntry.value = "";
              if (returnedText) {
                joditEditor.value += returnedText;
                closeStandardTextPopup();
              }
            })
            .catch(() => {
              hotkeyEntry.hidden = true;
              failureText.hidden = false;
              document.getElementById("standard_text_popup_ok_button").focus();
            });
        }
      });
      popup.hidden = false;
      hotkeyEntry.focus();
    }
  }
}

function closeStandardTextPopup() {
  // hide error messages and get ready for the next time its called
  const hotkeyEntry = document.getElementById("standard_text_popup_hotkey_entry");
  hotkeyEntry.value = "";
  hotkeyEntry.hidden = false;
  document.getElementById("standard_text_popup_failure").hidden = true;
  document.getElementById("standard_text_popup").hidden = true;
  joditEditor.editor.focus();
}
