/* global createLocalLNThread, createLocalSystemThread, log,
fixHtmlStandardText, fixPlaintext, oscObjects, allThreads, LNThread, SystemThread, TranslatedAgentThread, UntranslatedAgentThread */

/* eslint-disable prefer-promise-reject-errors */ /* we only want to wrap a string in an Error
object once to preserve the stack trace but also prevent multiple 'Error:' titles. */

/* eslint-disable-next-line no-unused-vars */ // used in other files
class OscRestManager {
  constructor(sessionToken, incidentId) {
    this._headers = {
      "Content-Type": "application/json",
      "OSvC-CREST-Application-Context": "Language I/O BUI Response REST Call",
      Authorization: "Session " + sessionToken,
    };
    this.incidentId = incidentId;
    this._baseUrl = oscObjects.globalContext.getInterfaceServiceUrl("REST") + "/connect/v1.4/";
    this._interfaceId = oscObjects.globalContext.getInterfaceId();
    this._accountNames = {};
    this._contactNames = {};
  }

  /* HELPERS */

  // wrapper for fetch to handle OSC errors, auth fails, fetch fails, and parsing fails consistently
  _safeFetch(fullUrl, method, body, secondAttempt = false, returnResponseDirectly = false, getSchema = false) {
    const parameters = {
      credentials: "same-origin",
      headers: this._headers,
      method,
    };

    if (getSchema) {
      parameters.headers.Accept = "application/schema+json";
    } else {
      parameters.headers.Accept = "*/*";
    }

    if (body) {
      parameters.body = JSON.stringify(body);
    }

    return new Promise((resolve, reject) => {
      fetch(fullUrl, parameters)
        /* eslint-disable-next-line consistent-return */ // we don't return an object for PATCHes
        .then((response) => {
          if (response.ok) {
            // OSC PATCHes don't return a json body, so just resolve with nothing
            if (method === "PATCH") {
              return resolve();
            } else if (returnResponseDirectly) {
              // for promise.all we need the response object
              return resolve(response);
            } else {
              // return the json body for the caller to process
              response
                .json()
                .then((responseJson) => {
                  if (responseJson) {
                    resolve(responseJson);
                  } else {
                    log.warn("Response body was empty.");
                    reject("Response body was empty.");
                  }
                })
                .catch(() => reject("Unable to parse OSC response body."));
            }
          } else {
            // if the call returned unauthorized, try again with a new session token - but only once
            /* eslint-disable-next-line no-lonely-if */ // does not apply
            if (response.status === 401 && !secondAttempt) {
              log.info("Unable to authenticate with OSC database. Trying again with a new session token.");
              oscObjects.globalContext
                .getSessionToken()
                .then((newSessionToken) => {
                  this._headers.Authorization = "Session " + newSessionToken;
                  return resolve(this._safeFetch(fullUrl, method, body, true));
                })
                .catch((error) => {
                  log.error(error);
                  return reject("Could not get new session token.");
                });
            } else {
              response
                .json()
                .then((errorJson) => {
                  log.warn(
                    new Error("Fetch failed: " + response.status + " " + response.statusText + "\n" + errorJson.detail)
                  );
                  reject("Could not communicate with OSC database.");
                })
                .catch(() => reject("Could not communicate with OSC database."));
            }
          }
        });
    });
  }

  static _getObjectsFromTable(json) {
    // format column and row lists into usable object list
    try {
      const fields = json.columnNames;
      const values = json.rows;
      const objects = [];
      let hasPopulatedValues = false;

      Object.keys(values).forEach((key) => {
        const obj = [];
        Object.keys(values[key]).forEach((innerKey) => {
          /* if there are no system threads, OSC will return a row array where every value is null,
          instead of just returning an empty row array like it does with LNThreads - so we need to
          check for this case and we will return an empty array if it occurs. */
          if (values[key][innerKey] !== null) {
            hasPopulatedValues = true;
            obj[fields[innerKey]] = values[key][innerKey];
          }
        });
        objects.push(obj);
      });
      if (hasPopulatedValues) {
        return objects;
      } else {
        return [];
      }
    } catch (error) {
      const modError = error;
      modError.message = "Could not parse objects: " + error.message;
      log.warn(modError);
      return null;
    }
  }

  /* GETTING LANGUAGE INFO */

  async getCustomLanguageLookupName(languageFieldDisplayName) {
    const fullUrl = this._baseUrl + "metadata-catalog/incidents/customFields/c/";
    return this._safeFetch(fullUrl, "GET", "", false, false, true)
      .then((json) => {
        let match;
        /* if we can't find a match below, we want to fall into the error case */
        /* eslint-disable-next-line consistent-return */
        Object.keys(json.properties).forEach((key) => {
          if (json.properties[key].label === languageFieldDisplayName) {
            match = key.toString();
          }
        });
        if (match) {
          return match;
        } else {
          throw new Error("No incident custom field with that name exists.");
        }
      })
      .catch((error) => {
        // add some context to the error message and re-throw
        const modError = error;
        modError.message = `Couldn't get custom language field '${languageFieldDisplayName}'. Verify that the 'languageField' server configuration property is set correctly for your profile, or contact your Language I/O Representative for assistance.`;
        throw modError;
      });
  }

  /* CREATING AND GETTING THREADS */

  createLNThread(
    text,
    threadType,
    sourceSystemThread,
    sourceLNThread,
    statusMessage,
    rnResponseProjectId,
    wordCount,
    translationId,
    ratingId = 0,
    ratingTypeId = 0,
    ratingPoints = 0,
    sourceLanguage,
    targetLanguage
  ) {
    const fullUrl = this._baseUrl + "linguistnow.LNThread";
    const body = {
      accountID: { id: oscObjects.globalContext.getAccountId() },
      contentType: 1, // always 1 for incident translation
      incidentID: { id: this.incidentId },
      ratingID: ratingId,
      ratingPoints,
      ratingTypeID: ratingTypeId,
      sourceLNProjectID: rnResponseProjectId,
      sourceLNThreadID: sourceLNThread,
      sourceThreadID: sourceSystemThread,
      statusMessage,
      text,
      threadType,
      translationID: translationId,
      wordCount,
      sourceLocale: sourceLanguage,
      targetLocale: targetLanguage
    };
    return this._safeFetch(fullUrl, "POST", body);
  }

  loadAllThreads() {
    const freshThreads = [];

    // get all of the ln threads
    const getLNThreads = this._getLNThreads().then((lnThreads) => {
      Object.keys(lnThreads).forEach((key) => {
        const ct = lnThreads[key];
        freshThreads.push(createLocalLNThread(ct));
      });
      return Promise.resolve();
    });

    // get all of the system threads
    const getSystemThreads = this._getSystemThreads().then((systemThreads) => {
      // get a list of all the new sender names we'll have to look up by their IDs
      const accounts = [];
      const contacts = [];

      Object.keys(systemThreads).forEach((key) => {
        if (systemThreads[key].account && !accounts.includes(systemThreads[key].account)) {
          accounts.push(systemThreads[key].account);
        } else if (systemThreads[key].contact && !contacts.includes(systemThreads[key].contact)) {
          contacts.push(systemThreads[key].contact);
        }
      });

      return this._getAllSystemThreadNames(accounts, contacts).then(() => {
        Object.keys(systemThreads).forEach((key) => {
          // now that we have all of the names, we can actually create the local threads
          const ct = systemThreads[key];
          let name = "";
          if (this._accountNames[ct.account]) {
            name = this._accountNames[ct.account];
          } else if (this._contactNames[ct.contact]) {
            name = this._contactNames[ct.contact];
          }
          freshThreads.push(createLocalSystemThread({ ...ct, name }));
        });
      });
    });

    return Promise.all([getLNThreads, getSystemThreads]).then(() => freshThreads);
  }

  _getLNThreads() {
    const fullUrl =
      this._baseUrl +
      "queryResults?query=SELECT ID, text, CreatedTime, threadType, sourceThreadID, sourceLNThreadID, statusMessage, translationID, ratingID, ratingTypeID, ratingPoints, sourceLocale, targetLocale FROM linguistnow.LNThread WHERE incidentID = " +
      this.incidentId;
    return new Promise((resolve, reject) => {
      this._safeFetch(fullUrl, "GET")
        .then((json) => resolve(OscRestManager._getObjectsFromTable(json.items[0])))
        .catch((error) => {
          const modError = error;
          modError.message = "Couldn't get LNThreads: " + error.message;
          log.warn(modError);
          return reject(modError);
        });
    });
  }

  _getSystemThreads() {
    const fullUrl =
      this._baseUrl +
      "queryResults?query=SELECT threads.id, threads.text, threads.createdTime, threads.contentType, threads.entryType, threads.account, threads.contact FROM incidents WHERE ID = " +
      this.incidentId;
    return new Promise((resolve, reject) => {
      this._safeFetch(fullUrl, "GET")
        .then((json) => resolve(OscRestManager._getObjectsFromTable(json.items[0])))
        .catch((error) => {
          const modError = error;
          modError.message = "Couldn't get System Threads: " + error.message;
          log.warn(modError);
          return reject(modError);
        });
    });
  }

  _getAllSystemThreadNames(accountIds, contactIds) {
    const promises = [];

    Object.keys(accountIds).forEach((key) => {
      promises.push(
        this._safeFetch(
          this._baseUrl + "queryResults?query=SELECT displayName FROM accounts WHERE id = " + accountIds[key],
          "GET"
        ).then((json) => {
          this._accountNames[accountIds[key]] = json.items[0].rows[0][0];
        })
      );
    });
    Object.keys(contactIds).forEach((key) => {
      promises.push(
        this._safeFetch(
          this._baseUrl + "queryResults?query=SELECT lookupName FROM contacts WHERE id = " + contactIds[key],
          "GET"
        ).then((json) => {
          this._contactNames[contactIds[key]] = json.items[0].rows[0][0];
        })
      );
    });

    return Promise.all(promises);
  }

  /* UPDATING THREAD/FIELD STATUSES */

  updateStatuses(lnStatusMessage, sourceLNThreadId, lnThreadStatusMessage) {
    this._updateLNStatus(lnStatusMessage);
    if (sourceLNThreadId > 0) {
      return this._updateLNThreadStatus(sourceLNThreadId, lnThreadStatusMessage);
    } else {
      return Promise.resolve();
    }
  }

  _updateLNThreadStatus(id, statusMessage) {
    const fullUrl = this._baseUrl + "linguistnow.LNThread/" + id;
    const body = {
      statusMessage,
    };
    return this._safeFetch(fullUrl, "PATCH", body).catch((error) => {
      const modError = error;
      modError.message = "Failed to update LNThread status: " + error.message;
      log.warn(modError);
    });
  }

  _updateLNStatus(statusMessage) {
    const fullUrl = this._baseUrl + "incidents/" + this.incidentId;
    const body = {
      customFields: {
        c: {
          ln_status: statusMessage,
        },
      },
    };
    oscObjects.extensionProvider.registerWorkspaceExtension((workspaceRecord) => {
      workspaceRecord.updateFieldByLabel('c$ln_status', statusMessage)
    });
    return this._safeFetch(fullUrl, "PATCH", body).catch((error) => {
      const modError = error;
      modError.message = "Failed to update Language I/O Status: " + error.message;
      log.warn(modError);
    });
  }

  _updateLNThreadRating(id, { ratingId = null, ratingTypeId = null, ratingPoints = null }) {
    const fullUrl = this._baseUrl + "linguistnow.LNThread/" + id;
    const body = {};
    if (ratingId !== null) {
      body.ratingID = ratingId;
    }
    if (ratingTypeId !== null) {
      body.ratingTypeID = ratingTypeId;
    }
    if (ratingPoints !== null) {
      body.ratingPoints = ratingPoints;
    }
    return this._safeFetch(fullUrl, "PATCH", body).catch((error) => {
      const modError = error;
      modError.message = "Failed to update LNThread rating: " + error.message;
      log.warn(modError);
    });
  }

  updateLanguageField(languageField, language){
    oscObjects.extensionProvider.registerWorkspaceExtension((workspaceRecord) => {
      workspaceRecord.updateField(`Incident.${languageField}`, language)
    });
  }
  /* GETTING STANDARD TEXT */

  getStandardText(hotkey) {
    return this._getStandardTextId(hotkey)
      .then((id) => this._getStandardTextValue(id))
      .catch((error) => {
        log.warn(error);
        return Promise.reject();
      });
  }

  _getStandardTextId(hotkey) {
    const fullUrl =
      this._baseUrl + "queryResults?query=SELECT id FROM StandardContents WHERE HotKey = '" + hotkey + "'";
    return this._safeFetch(fullUrl, "GET")
      .catch(() => {
        throw new Error("Couldn't get standard text: '" + hotkey + "' doesn't exist");
      })
      .then((json) => {
        if (json && json.status) {
          // we only get this field if its returning a query error
          throw new Error("Couldn't get standard text: " + json.status);
        } else {
          return json.items[0].rows[0][0]; // location of language value if successful
        }
      });
  }

  _getStandardTextValue(id) {
    const fullUrl = this._baseUrl + "standardContents/" + id + "/contentValues/";
    /* first, we try to get the HTML content. there may not be any, in which case we try to get the
     plain text content */
    return this._safeFetch(fullUrl + "2", "GET")
      .then((json) => fixHtmlStandardText(json.value))
      .catch(() => {
        log.warn("Couldn't get HTML value for standard content - attempting to get plaintext value instead.");
        return this._safeFetch(fullUrl + "1", "GET")
          .catch(() => {
            throw new Error(
              "Couldn't get standard text: id=" + id + " doesn't exist on interface=" + this._interfaceId
            );
          })
          .then((json) => {
            if (json && json.status) {
              // we only get this field if its returning a query error
              throw new Error("Couldn't get standard text: " + json.status);
            } else {
              return fixPlaintext(json.value);
            }
          });
      });
  }
}
