/* eslint-disable no-case-declarations */
/* global allThreads, log, fixPlaintext, fixHtmlThread, profileConfig */

/* eslint-disable class-methods-use-this */ /* the thread classes have some overridden functions
 where 'this' may or may not be used, so they shouldn't be static */
/* eslint-disable max-classes-per-file */ // including all thread subclasses since many are stubs
/* eslint-disable prefer-rest-params */ // we use 'arguments' when printing unexpected strings

class Thread {
  constructor({
    id,
    text,
    name,
    createdTime,
    threadType,
    sourceSystemThread,
    sourceLNThread,
    statusMessage,
    hierarchy,
    contentType,
  }) {
    this.id = parseInt(id, 10);
    this.name = name;
    this.createdTime = createdTime;
    this.timestamp = Thread._getFormattedDate(createdTime);
    this.threadType = parseInt(threadType, 10);
    this.sourceSystemThread = parseInt(sourceSystemThread, 10);
    this.sourceLNThread = parseInt(sourceLNThread, 10);
    this.contentType = parseInt(contentType, 10);

    // some system threads are plaintext and need extra processing for correct newline display
    let fixedText = text;
    if (this.contentType === 1) {
      fixedText = fixPlaintext(text);
    } else {
      fixedText = fixHtmlThread(text);
    }
    this.text = fixedText;

    this.statusType = "";
    this.statusMessage = statusMessage;
    // we're checking > 1 instead of > 0 because LNThread creation fails if the status is empty.
    if (statusMessage.length > 1) {
      this.statusType = "INFO";
      this.statusMessage = statusMessage;
      if (statusMessage.includes("SUCCESS")) {
        this.statusType = "SUCCESS";
      } else if (statusMessage.includes("ERROR")) {
        this.statusType = "ERROR";
      }
    }

    this.htmlId = this.constructor.name + this.id; // type + thread id used as unique id
    this.hierarchy = hierarchy; // whether the thread is a parent - used for display
  }

  getHtml() {
    return (
      '<div class="thread ' +
      this.hierarchy +
      '" id="' +
      this.htmlId +
      '">' +
      this._getStatusbar() +
      this._getContent() +
      "</div>"
    );
  }

  /* DISPLAY HELPERS */

  _getStatusbar() {
    let collapseButtonHtml = "";
    if (this.hierarchy === "parent") {
      collapseButtonHtml =
        '<button class="collapse_button icon" id="' + this.htmlId + 'collapse_button" + onclick=""></button>';
    }
    let statusHtml = "";
    if (this.statusMessage.length > 0) {
      // create expandable status with SUCCESS/INFO/ERROR 'button'
      statusHtml =
        '<button class="status type" onclick="toggleStatus(\'' +
        this.htmlId +
        "status_message')\">" +
        this.statusType +
        "</button>" +
        '<span class="status message collapsed" id="' +
        this.htmlId +
        'status_message">' +
        this.statusMessage +
        "</span>";
    }
    return (
      '<div class="statusbar ' +
      this.statusType +
      " " +
      this.constructor.name +
      '">' +
      collapseButtonHtml +
      this._getSenderIconHtml() +
      '<span class="name">' +
      this.name +
      "</span>" +
      '<span class="timestamp">' +
      this.timestamp +
      "</span>" +
      this._getButtonsHtml() +
      statusHtml +
      "</div>"
    );
  }

  static _getFormattedDate(oscDate) {
    const dt = new Date(oscDate);
    const month = (dt.getMonth() + 1).toString().padStart(2, "0");
    const date = dt.getDate().toString().padStart(2, "0");
    const year = dt.getFullYear();
    const rawHour = dt.getHours();
    const suffix = rawHour >= 12 ? "PM" : "AM";
    let hour = rawHour > 12 ? rawHour - 12 : rawHour;
    hour = hour === "00" ? 12 : hour;
    const minute = dt.getMinutes().toString().padStart(2, "0");
    const full = month + "/" + date + "/" + year + " " + hour + ":" + minute + " " + suffix;
    return full;
  }

  _getContent() {
    return '<div class="content" id="' + this.htmlId + 'content" dir="auto">' + this.text + "</div>";
  }

  /* DISPLAY HELPERS - OVERRIDDEN BY CHILD CLASSES */

  _getButtonsHtml() {
    return "";
  }

  _getSenderIconHtml() {
    return "";
  }
}

/* MAIN SYSTEM THREADS */

class SystemThread extends Thread {}

class NoteThread extends SystemThread {
  _getSenderIconHtml() {
    return '<img class="icon" src="./icons/senders/note.png" />';
  }
  _getButtonsHtml() {
    const buttonId = this.htmlId + "translate_button";
    return (
      '<button class="customer_translate_button" id="' +
      buttonId +
      '" onclick="translateContent(false, false, \'' +
      escape(this.text) +
      "', " +
      this.id +
      ", '" +
      buttonId +
      "', true);\">Translate</button>"
    );
  }
}

class SentAgentResponseThread extends SystemThread {
  _getSenderIconHtml() {
    return '<img class="icon" src="./icons/senders/saved-response.png" />';
  }
}

// this includes customer-created and proxy messages - we treat them the same
class UntranslatedCustomerThread extends SystemThread {
  _getButtonsHtml() {
    const buttonId = this.htmlId + "translate_button";
    return (
      '<button class="customer_translate_button" id="' +
      buttonId +
      '" onclick="translateContent(false, false, \'' +
      escape(this.text) +
      "', " +
      this.id +
      ", '" +
      buttonId +
      "');\">Translate</button>"
    );
  }

  _getSenderIconHtml() {
    return '<img class="icon" src="./icons/senders/customer.png" />';
  }
}

/* OTHER SYSTEM THREADS */

class ChatTranscriptThread extends SystemThread {
  _getSenderIconHtml() {
    return '<img class="icon" src="./icons/senders/chat.png" />';
  }
}

class AutomaticAgentResponseThread extends SystemThread {
  _getSenderIconHtml() {
    return '<img class="icon" src="./icons/senders/auto-response.png" />';
  }
}

class AutomaticNoteThread extends SystemThread {
  _getSenderIconHtml() {
    return '<img class="icon" src="./icons/senders/auto-note.png" />';
  }
}

// we're mostly going to treat this like a note, but we want to keep it distinct for debugging
class UnexpectedSystemThread extends SystemThread {
  _getSenderIconHtml() {
    return '<img class="icon" src="./icons/senders/note.png" />';
  }
}

/* LN THREADS */

class LNThread extends Thread {
  constructor({
    id,
    text,
    name,
    createdTime,
    threadType,
    sourceSystemThread,
    sourceLNThread,
    statusMessage,
    hierarchy,
    contentType,
    translationID,
    ratingID = 0,
    ratingTypeID = -1,
    ratingPoints = 0,
    sourceLocale,
    targetLocale
  }) {
    super({
      id,
      text,
      name,
      createdTime,
      threadType,
      sourceSystemThread,
      sourceLNThread,
      statusMessage,
      hierarchy,
      contentType,
    });
    this.translationID = translationID;

    this.agentRatingInstructions = profileConfig.qualityRatingInstructions;
    this.agentRatingTypeId = profileConfig.qualityRatingType;

    this.ratingId = parseInt(ratingID, 10);
    this.ratingTypeId = parseInt(ratingTypeID, 10);
    this.ratingPoints = parseInt(ratingPoints, 10);
    this.sourceLocale = sourceLocale;
    this.targetLocale = targetLocale;
  }

  /**
   * set info returned from the Language I/O server for a translation quality rating and re-render the message
   * @param {int} ratingId rating ID from the Language I/O server
   * @param {int} ratingPoints the points assigned to the rating
   */
  setRatingInfo(ratingId, ratingPoints) {
    this.ratingId = ratingId;
    this.ratingPoints = ratingPoints;

    document.getElementById(`translation_rating_${this.id}`).innerHTML = this.getRatingControlsHtml();
  }

  getHtml() {
    return (
      `<div class="thread ${this.hierarchy}" id="${this.htmlId}">` +
      this._getStatusbar() +
      '<div class="content_wrapper">' +
      this._getContent() +
      `<div id="translation_rating_${this.id}" class="quality_rating_controls">` +
      this.getRatingControlsHtml() +
      "</div>" +
      "</div>" +
      "</div>"
    );
  }

  getRatingControlsHtml() {
    if (!this.translationID && !this.ratingId) {
      // don't show the rating control(s) until there's something to rate
      return "";
    }

    if (this.ratingTypeId > 0 && this.agentRatingTypeId !== this.ratingTypeId) {
      // don't show the rating control(s) if the agent's configured rating type
      // does not match the rating type of an existing rating
      log.warn(
        `Agent's configured Quality Rating Type (${this.agentRatingTypeId}) does not match LNThread's Quality Rating Type (${this.ratingTypeId})`
      );
      return "";
    }

    switch (this.agentRatingTypeId) {
      case 1:
        const points = parseInt(this.ratingPoints, 10);
        // eslint-disable-next-line no-restricted-globals
        const ratingClass = isNaN(points) ? "rating_points_none" : "rating_points_" + points;
        const classes = "translation_quality_rating rating_type_flag " + ratingClass;
        const id = `translation_rating_${this.id}_flag`;
        const params = `${this.id}, ${this.translationID}, ${this.ratingId}, ${this.ratingPoints}`;
        return `<button class="${classes}" id="${id}" onClick="toggleTranslationFlagRating(${params})" title="${this.agentRatingInstructions}"></button>`;
      case 0:
      default:
        return "";
    }
  }
}

class UntranslatedAgentThread extends LNThread {
  getRatingControlsHtml() {
    // rating controls are not available for untranslated agent messages
    return "";
  }
}

class TranslatedAgentThread extends LNThread {
  _getButtonsHtml() {
    return (
      '<button class="agent_send_button" onclick="sendAgentInput(\'' +
      escape(this.text) +
      "');\">Send To Messages Tab</button>"
    );
  }

  getRatingControlsHtml() {
    // rating controls are not currently available for translations of agent messages
    return "";
  }
}

class TranslatedCustomerThread extends LNThread {
  _getButtonsHtml() {
    /* try to create a retranslate option - this requires finding the original untranslated thread
    and recording its text since we can't directly retranslate the thread text like usual */
    try {
      const result = allThreads.find((thread) => thread.id === this.sourceSystemThread);
      const originalText = result.text;

      const buttonId = this.htmlId + "retranslate_button";

      return (
        '<button class="customer_retranslate_button" id="' +
        buttonId +
        '" onclick="translateContent(false, true, \'' +
        escape(originalText) +
        "', " +
        this.sourceSystemThread +
        ", '" +
        buttonId +
        "');\">Retranslate</button>"
      );
    } catch (error) {
      log.info("Couldn't get parent text. Continuing without retranslate option: " + error);
      return "";
    }
  }
}

class RetranslatedCustomerThread extends LNThread {}

class PartiallyTranslatedCustomerThread extends LNThread {}

class UnexpectedLNThread extends LNThread {}

class TranslatedNoteThread extends LNThread {}

class PartiallyTranslatedNoteThread extends LNThread {}

/* THREAD FACTORIES */

// eslint-disable-next-line no-unused-vars
function createLocalSystemThread({ id, text, name, createdTime, entryType, contentType }) {
  // TODO: make these LNThread-only later
  const sourceSystemThread = 0;
  const sourceLNThread = 0;
  const statusMessage = 0;

  const threadData = {
    id,
    text,
    name,
    createdTime,
    threadType: entryType,
    sourceSystemThread,
    sourceLNThread,
    statusMessage,
    hierarchy: "parent",
    contentType,
  };

  const threadTypeInt = parseInt(entryType, 10);
  if (threadTypeInt === 1) {
    return new NoteThread(threadData);
  } else if (threadTypeInt === 2) {
    return new SentAgentResponseThread(threadData);
  } else if (threadTypeInt === 3 || threadTypeInt === 4) {
    return new UntranslatedCustomerThread(threadData);
  } else if (threadTypeInt === 5) {
    return new ChatTranscriptThread(threadData);
  } else if (threadTypeInt === 6) {
    return new AutomaticAgentResponseThread(threadData);
  } else if (threadTypeInt === 7) {
    return new AutomaticNoteThread(threadData);
  } else {
    log.warn(new Error("We encountered an unexpected system thread: " + JSON.stringify(arguments)));
    return new UnexpectedSystemThread(threadData);
  }
}

// eslint-disable-next-line no-unused-vars
function createLocalLNThread({
  id,
  text,
  createdTime,
  threadType,
  sourceThreadID,
  sourceLNThreadID,
  statusMessage,
  translationID,
  ratingID = 0,
  ratingTypeID = -1,
  ratingPoints = 0,
  sourceLocale,
  targetLocale
}) {
  const contentType = 2; // LNThreads are always HTML so we want to avoid extra format processing

  const threadData = {
    id,
    text,
    name: "",
    createdTime,
    threadType,
    sourceSystemThread: sourceThreadID,
    sourceLNThread: sourceLNThreadID,
    statusMessage,
    hierarchy: "",
    contentType,
    translationID,
    ratingID,
    ratingTypeID,
    ratingPoints,
    sourceLocale,
    targetLocale
  };

  const threadTypeInt = parseInt(threadType, 10);
  if (threadTypeInt === 1) {
    return new UntranslatedAgentThread({ ...threadData, name: "Untranslated Agent Thread", hierarchy: "parent" });
  } else if (threadTypeInt === 2) {
    return new TranslatedAgentThread({ ...threadData, name: "Agent Thread Translation", hierarchy: "child" });
  } else if (threadTypeInt === 4) {
    return new TranslatedCustomerThread({ ...threadData, name: "Customer Thread Translation", hierarchy: "child" });
  } else if (threadTypeInt === 5) {
    return new RetranslatedCustomerThread({ ...threadData, name: "Customer Thread Retranslation", hierarchy: "child" });
  } else if (threadTypeInt === 6) {
    return new PartiallyTranslatedCustomerThread({
      ...threadData,
      name: "Customer Thread Partial Translation",
      hierarchy: "child",
    });
  } else if (threadTypeInt === 7){
      return new TranslatedNoteThread({...threadData, name: "Private Note Translation", hierarchy: "child"})
  } else if (threadTypeInt === 8){
    return new PartiallyTranslatedNoteThread({...threadData, name: "Private Note Partial Translation", hierarchy: "child"})
  } else {
    log.warn(new Error("We encountered an unexpected LN thread: " + JSON.stringify(arguments)));
    return new UnexpectedLNThread({ ...threadData, name: "Internal Language I/O Thread", hierarchy: "parent" });
  }
}
